/*
	This file is part of FaMaTS.

    FaMaTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FaMaTS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with FaMaTS.  If not, see <http://www.gnu.org/licenses/>.

 */
package es.us.isa.JaCoPReasoner;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import JaCoP.constraints.And;
import JaCoP.constraints.Constraint;
import JaCoP.constraints.IfThen;
import JaCoP.constraints.IfThenElse;
import JaCoP.constraints.In;
import JaCoP.constraints.PrimitiveConstraint;
import JaCoP.constraints.Sum;
import JaCoP.constraints.XeqC;
import JaCoP.constraints.XeqY;
import JaCoP.constraints.XgtC;
import JaCoP.core.Domain;
import JaCoP.core.FDV;
import JaCoP.core.FDstore;
import JaCoP.core.IntervalDomain;
import JaCoP.core.Variable;
import JaCoP.search.ComparatorVariable;
import JaCoP.search.LargestDomain;
import JaCoP.search.LargestMin;
import JaCoP.search.MaxRegret;
import JaCoP.search.MinDomainOverDegree;
import JaCoP.search.MostConstrainedDynamic;
import JaCoP.search.MostConstrainedStatic;
import JaCoP.search.SmallestDomain;
import JaCoP.search.SmallestMax;
import JaCoP.search.SmallestMin;



import es.us.isa.FAMA.Benchmarking.PerformanceResult;

import es.us.isa.FAMA.Exceptions.FAMAException;
import es.us.isa.FAMA.Exceptions.FAMAParameterException;
import es.us.isa.FAMA.Reasoner.FeatureModelReasoner;
import es.us.isa.FAMA.Reasoner.Question;
import es.us.isa.FAMA.models.featureModel.Cardinality;
import es.us.isa.FAMA.models.featureModel.GenericFeature;
import es.us.isa.FAMA.models.featureModel.GenericRelation;

public class JaCoPReasoner extends FeatureModelReasoner {
	private Map<String,GenericFeature> features;
	private GenericFeature root;
	private FDstore store;
	private Map<String,FDV> variables;
	private ArrayList<Constraint> constraints;
	private Map<Constraint,GenericRelation> relationConstraintMap;
	protected boolean consistent;
	protected boolean reify;
	protected ComparatorVariable heuristics;
	protected static Map<String,ComparatorVariable> heuristicsMap;
	
	public JaCoPReasoner() {
		store = new FDstore();
		variables = new HashMap<String,FDV>();
		constraints = new ArrayList<Constraint>();
		relationConstraintMap = new HashMap<Constraint,GenericRelation>();
		consistent = true;
		reify = false;
		heuristics = null;
		heuristicsMap = new HashMap<String,ComparatorVariable>();
		heuristicsMap.put("MCD",new MostConstrainedDynamic());
		heuristicsMap.put("MCS",new MostConstrainedStatic());
		heuristicsMap.put("LD",new LargestDomain());
		heuristicsMap.put("LM",new LargestMin());
		heuristicsMap.put("MR",new MaxRegret());
		heuristicsMap.put("MDOD",new MinDomainOverDegree());
		heuristicsMap.put("SDOM",new SmallestDomain());
		heuristicsMap.put("SMAX",new SmallestMax());
		heuristicsMap.put("SMIN",new SmallestMin());
		features = new HashMap<String,GenericFeature>();
	}
	
	@Override
	public void reset() {
		store = new FDstore();
		variables = new HashMap<String,FDV>();
		constraints = new ArrayList<Constraint>();
		relationConstraintMap = new HashMap<Constraint,GenericRelation>();
		consistent = true;
		reify = false;
		root = null;
		features = new HashMap<String,GenericFeature>();
	}
	
	@Override
	public void addFeature(GenericFeature f, Collection<Cardinality> cards) {
		features.put(f.getName(), f);
		Iterator<Cardinality> cardIt = cards.iterator();
		FDV var = new FDV(store,f.getName(),0,1);
		while (cardIt.hasNext()) {
			Cardinality card = cardIt.next();
			var.addDom(card.getMin(),card.getMax());
		}
		variables.put(f.getName(),var);
	}

	@Override
	public void addRoot(GenericFeature feature) {
		Variable root = variables.get(feature.getName());
		PrimitiveConstraint constraint = new XeqC(root,1);
		constraints.add(constraint);
		store.impose(constraint);
		//relationConstraintMap.put(null,constraint);
	}

	@Override
	public void addMandatory(GenericRelation rel,GenericFeature child, GenericFeature parent) {
		Variable childVar  = variables.get(child.getName());
		Variable parentVar = variables.get(parent.getName());
		PrimitiveConstraint constraint = new XeqY(childVar,parentVar);
		constraints.add(constraint);
		store.impose(constraint);
		relationConstraintMap.put(constraint,rel);
	}

	@Override
	public void addOptional(GenericRelation rel,GenericFeature child, GenericFeature parent) {
		Variable childVar  = variables.get(child.getName());
		Variable parentVar = variables.get(parent.getName());
		PrimitiveConstraint constraint = new IfThen(new XeqC(parentVar,0),new XeqC(childVar,0));;
		constraints.add(constraint);
		store.impose(constraint);
		relationConstraintMap.put(constraint,rel);
	}

	@Override
	public void addCardinality(GenericRelation rel,GenericFeature child, GenericFeature parent,
			Iterator<Cardinality> cardinalities) {
		Variable childVar  = variables.get(child.getName());
		Variable parentVar = variables.get(parent.getName());
		
		Domain domain = new IntervalDomain() ;
		if (!cardinalities.hasNext())
			domain.addDom(0, 0);
		while(cardinalities.hasNext()) {
			Cardinality card = cardinalities.next();
			domain.addDom(card.getMin(),card.getMax());
		}
		
		//FIXME no s� si funcionar� correctamente esta nueva expresi�n, creo que s�
		PrimitiveConstraint constraint = new IfThenElse(new XeqC(parentVar,0),new XeqC(childVar,0),new In(childVar,domain));
		constraints.add(constraint);
		store.impose(constraint);
		relationConstraintMap.put(constraint,rel);
		//TODO constraint done 4 David Benavides
		PrimitiveConstraint constraintG = new IfThen(new XeqC(childVar,0),new XeqC(parentVar,0));
		constraints.add(constraintG);
		store.impose(constraintG);
		relationConstraintMap.put(constraintG,rel);
	}

	@Override
	public void addSet(GenericRelation rel,GenericFeature parent, Collection<GenericFeature> children,
			Collection<Cardinality> cardinalities) {
		FDV parentVar = variables.get(parent.getName());
		// create the sum variable
		FDV setVar = new FDV(store,rel.getName());
		setVar.addDom(0,children.size());
		Domain domain= new IntervalDomain();
		Iterator<Cardinality> itc = cardinalities.iterator();
		while(itc.hasNext()) {
			Cardinality card = itc.next();
			domain.addDom(card.getMin(),card.getMax());
		}
		variables.put(rel.getName(),setVar);
		
		// create a set of variables to sum them
		ArrayList<Variable> varsList = new ArrayList<Variable>();
		Iterator<GenericFeature> it = children.iterator();
		while (it.hasNext()) {
			varsList.add(variables.get(it.next().getName()));
		}
		
		// creates the sum constraint with the cardinality variable
		Constraint suma = new Sum(varsList, setVar);
		// it is not a primitive constraint and cannot be reified
		constraints.add(suma);
		store.impose(suma);
		
		// creates the arraylist of constraints for the And constraint
		ArrayList<PrimitiveConstraint> localConstraints = new ArrayList<PrimitiveConstraint>();
		it = children.iterator();
		while (it.hasNext()) {
			Variable childVar = variables.get(it.next().getName());
			localConstraints.add(new XeqC(childVar, 0));
		}
		
		// adding the second constraint to the store
		//TODO cambiar In por XeqC(setVar,0)
		PrimitiveConstraint constraint = new IfThenElse(
				new XgtC(parentVar, 0), new In(setVar, domain),
				//new In(setVar,new FD(0,0)));
				new And(localConstraints));
		constraints.add(constraint);
		store.impose(constraint);
		relationConstraintMap.put(constraint,rel);
		
		//TODO done 4 David Benavides
		PrimitiveConstraint constraintG = new IfThen(new XeqC(setVar,0),new XeqC(parentVar,0));
		constraints.add(constraintG);
		store.impose(constraintG);
		relationConstraintMap.put(constraintG,rel);
	}

	@Override
	public void addExcludes(GenericRelation rel,GenericFeature origin, GenericFeature destination) {
		Variable originVar  = variables.get(origin.getName());
		Variable destVar = variables.get(destination.getName());
		//FIXME it should work in this way, but not sure
		PrimitiveConstraint constraint = new IfThen(new XgtC(originVar,0),new XeqC(destVar,0));
		constraints.add(constraint);
		store.impose(constraint);
		relationConstraintMap.put(constraint,rel);
	}

	@Override
	public void addRequires(GenericRelation rel,GenericFeature origin, GenericFeature destination) {
		Variable originVar  = variables.get(origin.getName());
		Variable destVar = variables.get(destination.getName());
		//FIXME it should work in this way, but not sure
		PrimitiveConstraint constraint = new IfThen(new XgtC(originVar,0),new XgtC(destVar,0));
		constraints.add(constraint);
		store.impose(constraint);
		relationConstraintMap.put(constraint,rel);
	}

	public PerformanceResult ask(Question q) throws FAMAException {
		if(q==null){
			throw new FAMAParameterException("Question: Not specified");
		}
		PerformanceResult res;
		JaCoPQuestion jq = (JaCoPQuestion)q;
		if (heuristics != null)
			jq.setHeuristics(heuristics);
		jq.preAnswer(this);
		res = jq.answer(this);
		jq.postAnswer(this);
		consistent = true;
		return res;
	}
	
	public void setReify (boolean reify) {
		this.reify = reify;
	}
	
	GenericRelation getRelation(Constraint c) {
		return relationConstraintMap.get(c);
	}
	
	public boolean consistency() {
		if (consistent)
			consistent = store.consistency();
		return consistent;
	}
	
	public ArrayList<FDV> getVariables() {
		ArrayList<FDV> res = new ArrayList<FDV>();
		Iterator<Entry<String,FDV>> it = variables.entrySet().iterator();
		while (it.hasNext()) {
			Entry<String,FDV> elem = it.next();
			res.add(elem.getValue());
		}
		return res;
	}
	
	public FDstore getStore(){
		return store;
	}

	public GenericFeature searchFeatureByName(String id) {
		return features.get(id);
	}
	
	public GenericFeature getRoot() {
		return root;
	}
	
	public Collection<GenericFeature> getAllFeatures(){
		return this.features.values();
	}
}
